package com.dashrava.utils {
	/**
	 * ...
	 * @author Orava
	 */
	public class GraphicUtils {

		/**
		 * Scale all components of a hex.
		 * @param	hex
		 * @param	scale
		 * @return	Scaled hex
		 */

		public static function scalehex(hex:uint, scale:Number):uint {

			var r:int = (hex >> 16	) & 255;
			var g:int = (hex >> 8	) & 255;
			var b:int = (hex		) & 255;

			r = clampInt(r * scale);
			g = clampInt(g * scale);
			b = clampInt(b * scale);

			var col:uint = r << 16 | g << 8 | b;

			return col;

		}

		public static function clampInt (val:Number, min:int = 0, max:int = 255):int {

			if (val < min) return min;
			if (val > max) return max;
			return int(val);
		}

		/**
		 * Get brightness/luminance of hex
		 * @param	hex
		 * @return
		 */
		public static function hexBrightness(hex:uint):Number {
			var rgb:Array = hexToRGB(hex);

			var R:int = rgb[0];
			var G:int = rgb[1];
			var B:int = rgb[2];

			return 0.241 * (R * R) + 0.691 * (G * G) + 0.068 * (B * B);
		}

		/**
		 * Converts hex to RGB
		 * @param	hex
		 * @return Array [r,g,b]
		 */

		public static function hexToRGB(hex:uint):Array {
			var rgb:Array = new Array();
			rgb[0] = hex >> 16;
			rgb[1] = hex >> 8 & 0xFF;
			rgb[2] = hex & 0xFF;

			if (rgb[0] < 0) { // Fix 32-bit
				rgb[0] = rgb[0] + 256;
			}

			return rgb;
		}

		/**
		 * Converts RGB to hex
		 * @param	red
		 * @param	green
		 * @param	blue
		 * @return hex uint
		 */
		public static function RGBToHex(r:Number, g:Number, b:Number):uint {
			var sr:String = r.toString(16);
			var sg:String = g.toString(16);
			var sb:String = b.toString(16);

			// Flash has a nasty habit of messing up zeroes in uints
			if (sr.length == 1) { sr = "0" + r; }
			if (sg.length == 1) { sg = "0" + g; }
			if (sb.length == 1) { sb = "0" + b; }
			return uint("0x"+(sr+sg+sb).toUpperCase());
		}

	}

}