/**
 * Generate a fantasy-world map
 * License: MIT
 * via https://github.com/amitp/mapgen
 * @author amitp (amitp@cs.stanford.edu), original creator.
 * @author dashrava (dashrava.com), modified by / converting from terrain map to planet gen.
 */

package {

	import com.dashrava.utils.GraphicUtils;
	import com.dashrava.utils.Random;
	import flash.geom.*;
	import flash.display.*;
	import flash.text.*;
	import flash.events.*;
	import flash.utils.*;
	import flash.net.*;

	public class planetgen extends Sprite {

		public var light:Point = new Point();
		public var pixelScale:int = 4;
		public var planetname:String;

		public var SEED:int = 42;
		public var OCEAN_ALTITUDE:int = 1;
		public var SIZE:int = 64;

		public var map:Map = new Map(SIZE, SEED);

		public var detailMap:Sprite;
		public var overlayMap:Sprite;
		public var starMap:Sprite;
		public var atmos:Sprite;

		public var COL_SEA:uint = 0x000099;
		public var COL_SHORE:uint = 0xae8c4c;
		public var COL_SNOW:uint = 0xffffff;
		public var COL_WET:uint = 0x56821b;
		public var COL_MOIST:uint = 0x3b8c43;
		public var COL_NORMAL:uint = 0x54653c;
		public var COL_PARCHED:uint = 0x334021;
		public var COL_DRY:uint = 0x989a2d;
		public var COL_DEAD:uint = 0xc2bd8c;

		public var smoothEnabled:Boolean = true;		// Tile smoothing method using corner adjusting magic
		public var planetmaskEnabled:Boolean = true; 	// Round the planet tilemap with a mask
		public var atmosphereEnabled:Boolean = true;
		public var starfieldEnabled:Boolean = true;
		public var cloudsEnabled:Boolean = true;
		public var debrisEnabled:Boolean = true;
		public var bandedEnabled:Boolean = true;		// Use banded light instead of a full gradient?

		// See Main for suggested ranges. Some of this is super arbitrary.
		public var cloudDensity:Number = 0.2;
		public var debrisDensity:Number = 0.2;
		public var debrisGap:Number = 0.2;
		public var debrisField:Number = 0.2;
		public var lightRadius:Number = 0.5;
		public var atmosIntensity:Number = 0.3;
		public var atmosRadius:Number = 1.1;
		public var lightIntensity:Number = 1;
		public var lightMinimum:Number = 0.5;

		public function planetgen() {

			roll();
		}

		public function roll():void {

			Random.seed = SEED;

			planetname = generateName();

			SIZE = Random.value() * 64 + 32;

			COL_SEA = randomColour();
			COL_SHORE = randomColour();
			COL_SNOW = randomColour();
			COL_WET = randomColour();
			COL_MOIST = randomColour();
			COL_NORMAL = randomColour();
			COL_PARCHED = randomColour();
			COL_DRY = randomColour();
			COL_DEAD = randomColour();

			map = new Map(SIZE, SEED);
			map.generate();

			if (detailMap 	&& contains(detailMap)) 	removeChild(detailMap);
			if (starMap 	&& contains(starMap)) 		removeChild(starMap);
			if (overlayMap 	&& contains(overlayMap)) 	removeChild(overlayMap);

			detailMap = new Sprite();
			overlayMap = new Sprite();
			starMap = new Sprite();

			// Starfield
			generateStars();

			starMap.cacheAsBitmap = true;

			addChild(starMap);

			starMap.visible = starfieldEnabled;

			if (atmos) {
				removeChild(atmos);
			}

			generateDetailMap(detailMap.graphics, overlayMap.graphics);

			var atmoSize:Number = detailMap.width * 0.5 * atmosRadius;

			var mat:Matrix = new Matrix();
			mat.createGradientBox(SIZE * lightRadius * 2, SIZE * lightRadius * 2, 0, light.x, light.y);

			atmos = new Sprite();

			var ratios:Array = [50 + Random.value() * Random.value() * 150, 255]
			var alphas:Array = [0.1 + atmosIntensity, atmosIntensity * 0.5 + 0.05];

			if (atmosphereEnabled) {
				atmos.graphics.beginGradientFill(GradientType.RADIAL, [COL_SEA, GraphicUtils.scalehex(COL_SEA, 0.6)], alphas, ratios, mat);
				atmos.graphics.drawCircle(SIZE * 0.5, SIZE * 0.5, atmoSize);
			}

			if (planetmaskEnabled) {
				var pmask:Sprite = new Sprite();
				pmask.name = "planetmask";
				pmask.graphics.lineStyle(1, 0xFFFFFF);
				pmask.graphics.beginFill(0xFF0000, 0.1);
				pmask.graphics.drawCircle(SIZE * 0.5, SIZE * 0.5, detailMap.width * 0.45);

				detailMap.addChild(pmask);
				detailMap.mask = pmask;
			}

			detailMap.scaleX = detailMap.scaleY 	= pixelScale;
			overlayMap.scaleX = overlayMap.scaleY 	= pixelScale;
			atmos.scaleX = atmos.scaleY 			= pixelScale;

			addChild(detailMap);
			addChild(overlayMap);
			addChild(atmos);

			detailMap.x 	= detailMap.y 	= -SIZE*0.5*pixelScale;
			overlayMap.x 	= overlayMap.y 	= -SIZE*0.5*pixelScale;
			atmos.x 		= atmos.y 		= -SIZE*0.5*pixelScale;

		}

		private function generateStars():void {

			var stars = Random.value() * 500 + 100;

			var scol:uint;

			var starx:Number;
			var stary:Number;
			var starr:Number;

			for (var j:int = 0; j < stars; j++) {

				var rR:Number = 90 + Math.random() * 50;
				var rB:Number = rR - Math.random() * 50;
				var rG:Number = rR;

				scol = GraphicUtils.RGBToHex(rR, rG, rB);

				starx = Random.value() * 1280 - 1280 * 0.5;
				stary = Random.value() * 720  -  720 * 0.5;
				starr = Math.random() * Math.random() * 4;

				starMap.graphics.beginFill(scol, Math.random());

				if (smoothEnabled) {
					starMap.graphics.drawCircle(starx, stary, starr);
				}
				else {
					starMap.graphics.drawRect(starx, stary, starr, starr);
				}
			}
		}

		private function generateName():String {

			var chars:Array = "aeiouyaeiouyaeiouyabcdefghijklmnopqrstuvxyz".split("");

			var chipname:String = "";

			// Make a random string
			var l:int = int(4 + Random.value() * Random.value() * 20);

			for (var i:int = 0; i < l; i++) {
				chipname += chars[int(Random.value() * chars.length)];
			}

			// Add a hyphen
			if (Random.value() > 0.8) {
				var sub:int = (Random.value() * chipname.length-2) >> 0;
				sub += 2;
				chipname = chipname.substr(0, sub) + "-" + chipname.substr(sub);
			}

			// Add a number
			if (Random.value() > 0.8) {
				chipname = chipname + " " + String((Random.value() * Random.value() * 999) >> 0);
			}

			return chipname;
		}

		function randomColour():uint {
			var rR:Number = 60 + Random.value() * 120;
			var rG:Number = 60 + Random.value() * 120;
			var rB:Number = 60 + Random.value() * 120;

			return GraphicUtils.RGBToHex(rR, rG, rB);
		}

		public function moistureAndAltitudeToColor(m:Number, a:Number):int {
			var color:int = 0xff0000;

			if (a < OCEAN_ALTITUDE) {
				color = COL_SEA;
			} else if (a < OCEAN_ALTITUDE + 5) {
				color = GraphicUtils.scalehex(COL_WET, 0.8);
			} else if (m > 200)
				color = COL_WET;
			else if (m > 150)
				color = GraphicUtils.scalehex(COL_WET, 0.9);
			else if (m > 100)
				color = GraphicUtils.scalehex(COL_WET, 0.8);
			else if (m > 50)
				color = GraphicUtils.scalehex(COL_NORMAL, 1.1);
			else if (m > 20)
				color = GraphicUtils.scalehex(COL_NORMAL, 1.2);
			else
				color = GraphicUtils.scalehex(COL_NORMAL, 1.3);

			/*else if (a > 200) {
			   color = GraphicUtils.scalehex(COL_SNOW, a / 255);
			 }*/

			return color;
		}

		private function generateDetailMap(planet:Graphics, overlay:Graphics):void {

			var centerX:Number = SIZE * 0.5;
			var centerY:Number = SIZE * 0.5;

			// Parameters
			var cornerAdjust:Number = 0.2;
			var randomAdjust:Number = 0;

			if (smoothEnabled) {
				cornerAdjust = 0.25;
			}
			else {
				cornerAdjust = 0;
			}

			// We are drawing an area DETAILSIZE x DETAILSIZE.
			var x:int, y:int;
			planet.clear();
			overlay.clear();

			// Coordinates of the detail area:
			var bounds:Rectangle = new Rectangle(0, 0, SIZE, SIZE);

			// Make sure that we're entirely within the bounds of the map
			//bounds = bounds.intersection(new Rectangle(0, 0, SIZE, SIZE));

			// 2d Array of vertices
			var vertices:Array = [];
			for (x = bounds.left; x <= bounds.right; x++) {
				vertices[x] = [];
				for (y = bounds.top; y <= bounds.bottom; y++) {
					// TODO: we'd save a lot of allocation if we reused these points
					vertices[x][y] = new Point(x, y);
				}
			}

			// Move vertices randomly
			if (randomAdjust > 0) {
				var noise:BitmapData = new BitmapData(256, 256);
				noise.noise(SEED, 0, 255);
				for (x = bounds.left; x <= bounds.right; x++) {
					for (y = bounds.top; y <= bounds.bottom; y++) {

						var noiseColor:int = noise.getPixel(x % 256, y % 256);
						var rand1:Number = ((noiseColor & 0x00ff00) >> 8) / 255.0 - 0.5;
						var rand2:Number = (noiseColor & 0xff) / 255.0 - 0.5;
						vertices[x][y].x += rand1 * randomAdjust;
						vertices[x][y].y += rand2 * randomAdjust;
					}
				}
			}

			if (cornerAdjust > 0) {
				// Alter vertices if 3 of 4 squares has same type
				for (x = bounds.left + 1; x < bounds.right; x++) {
					for (y = bounds.top + 1; y < bounds.bottom; y++) {

						// Sprites at the four squares touching this vertex
						var cTL:int = moistureAndAltitudeToColor(map.moisture[x - 1][y - 1], map.altitude[x - 1][y - 1]);
						var cTR:int = moistureAndAltitudeToColor(map.moisture[x][y - 1], map.altitude[x][y - 1]);
						var cBL:int = moistureAndAltitudeToColor(map.moisture[x - 1][y], map.altitude[x - 1][y]);
						var cBR:int = moistureAndAltitudeToColor(map.moisture[x][y], map.altitude[x][y]);

						// Figure out which corner is odd, if any
						if (cTR == cBL && cBL == cBR && cTL != cTR) { // TL is odd
							vertices[x][y].x -= cornerAdjust;
							vertices[x][y].y -= cornerAdjust;
						}
						if (cTL == cTR && cTR == cBL && cBL != cBR) { // BR is odd
							vertices[x][y].x += cornerAdjust;
							vertices[x][y].y += cornerAdjust;
						}
						if (cTL == cBL && cBL == cBR && cTL != cTR) { // TR is odd
							vertices[x][y].x += cornerAdjust;
							vertices[x][y].y -= cornerAdjust;
						}
						if (cTL == cTR && cTR == cBR && cTL != cBL) { // BL is odd
							vertices[x][y].x -= cornerAdjust;
							vertices[x][y].y += cornerAdjust;
						}
							// TODO: this seems like stupid repetitive code TODO: what
							// should we do when we have tiles A A B C (A and A are
							// adjacent but only two of them)?
					}
				}
			}

			// Draw grid
			var g:Graphics;

			//g.lineStyle(1, 0x000000, 0.1);   // uncomment for tile borders

			var m:Matrix = new Matrix();

			for (x = bounds.left; x < bounds.right; x++) {
				for (y = bounds.top; y < bounds.bottom; y++) {

					var vx:Number = vertices[x][y].x;
					var vy:Number = vertices[x][y].y;

					var dx:Number = (vx - SIZE * 0.5) * (vx - SIZE * 0.5);
					var dy:Number = (vy - SIZE * 0.5) * (vy - SIZE * 0.5);

					if (dx + dy > SIZE * 0.50 * SIZE * 0.50) {
						continue; // don't draw stuff in the corners
					}

					// Separating zone between planet and debris
					if (dx + dy > SIZE * debrisField * SIZE * debrisField) {

						if (dx + dy < SIZE * (debrisField+debrisGap) * SIZE * (debrisField+debrisGap)) {
							continue;
						}
					}

					// debris density
					if (dx + dy > SIZE * debrisField * SIZE * debrisField) {

						if (!debrisEnabled) continue;

						if (Math.random() > debrisDensity) {
							continue;
						}
					}

					// TODO: we should have cached the tile ids when we generate the map...
					var c:int = moistureAndAltitudeToColor(map.moisture[x][y], map.altitude[x][y]);

					var scal:Number = ((SIZE * lightRadius * SIZE * lightRadius) / ((vx - light.x - SIZE * lightRadius) * (vx - light.x - SIZE * lightRadius) + (vy - light.y - SIZE * lightRadius) * (vy - light.y - SIZE * lightRadius)));

					// Light banding because cool
					if (bandedEnabled) {
						if (scal > 1.2) {
							scal = 1.2;
						}
						else if (scal > 1) {
							scal = 1;
						}
						else if (scal > 0.8) {
							scal = 0.8;
						}
						else if (scal > 0.6) {
							scal = 0.6;
						}
						else {
							scal = lightMinimum;
						}
					}
					else {
						if (scal > 1.2) {
							scal = 1.2;
						}
						else if (scal < lightMinimum) {
							scal = lightMinimum;
						}
					}

					//if (scal != lightMinimum) {
						scal *= lightIntensity;
					//}

					c = GraphicUtils.scalehex(c, scal);

					var debris:Boolean = (dx + dy > SIZE * debrisField * SIZE * debrisField);
					var cloud:Boolean = (Math.random() < cloudDensity);

					if ((debrisEnabled && debris)) {

						g = overlay;

						c = GraphicUtils.scalehex(Math.random()>0.5?COL_NORMAL:COL_WET, scal * (0.75 + Math.random() * 0.25));
					}
					else if ((cloudsEnabled && cloud)) {
						g = overlay;

						c = GraphicUtils.scalehex(COL_SEA, scal * (1.25 + Math.random() * 0.25));
					}
					else {
						g = planet;
					}

					g.beginFill(c);
					g.moveTo(vertices[x][y].x, vertices[x][y].y);
					g.lineTo(vertices[x][y + 1].x, vertices[x][y + 1].y);
					g.lineTo(vertices[x + 1][y + 1].x, vertices[x + 1][y + 1].y);
					g.lineTo(vertices[x + 1][y].x, vertices[x + 1][y].y);
					g.lineTo(vertices[x][y].x, vertices[x][y].y);
					g.endFill();
				}
			}

			g.lineStyle();

		}

	}
}

import com.dashrava.utils.Random;
import flash.display.*;
import flash.geom.*;

class Map {
	// Make the map into an island.  0.39 makes almost every map into an
	// island; 0.1 makes almost no map into an island.
	public var ISLAND_EFFECT:Number = 0.01;

	public var SIZE:int;
	public var SEED:int;

	public var altitude:Vector.<Vector.<int>>;
	public var moisture:Vector.<Vector.<int>>;

	function Map(size:int, seed:int) {
		SIZE = size;
		SEED = seed;
		altitude = make2dArray(SIZE, SIZE);
		moisture = make2dArray(SIZE, SIZE);
	}

	public function generate():void {
		// Generate 3-channel perlin noise and copy 2 of the channels out
		var b:BitmapData = new BitmapData(SIZE, SIZE);
		b.perlinNoise(SIZE, SIZE, 4, SEED, false, false);

		var s:Shape = new Shape();

		// NOTE: if we remembered the equalization and other parameters,
		// we could zoom in at least on the altitude map by using the
		// perlinNoise() size and offsets parameters to regenerate
		// portions of the map at higher resolution. This could be useful
		// when drawing the detail map.
		//equalizeTerrain(b);

		// Overlay this on an "island" multiplier map
		// Based on http://www.ridgenet.net/~jslayton/FunWithWilburVol6/index.html
		for (x = 0; x < SIZE; x++) {
			for (y = 0; y < SIZE; y++) {
				var radiusSquared:Number = (x - SIZE * 0.5) * (x - SIZE * 0.5) + (y - SIZE * 0.5) * (y - SIZE * 0.5);
				radiusSquared += Math.pow(Math.max(Math.abs(x - SIZE * 0.5), Math.abs(y - SIZE * 0.5)), 2);
				var max_radiusSquared:Number = SIZE * SIZE / 4;
				radiusSquared /= max_radiusSquared;
				radiusSquared += Random.value() * 0.1;
				radiusSquared *= ISLAND_EFFECT;
				var island_multiplier:Number = Math.exp(-radiusSquared / 4) - radiusSquared;
				island_multiplier = island_multiplier * island_multiplier * island_multiplier;

				c = b.getPixel(x, y);
				var height:int = (c >> 8) & 0xff;
				height = int(island_multiplier * height);
				if (height < 0)
					height = 0;
				if (height > 255)
					height = 255;
				c = (c & 0xffff00ff) | (height << 8);
				b.setPixel(x, y, c);
			}
		}

		equalizeTerrain(b);

		// Extract information from bitmap
		for (var x:int = 0; x < SIZE; x++) {
			for (var y:int = 0; y < SIZE; y++) {
				var c:int = b.getPixel(x, y);
				altitude[x][y] = (c >> 8) & 0xff;
				moisture[x][y] = c & 0xff;
			}
		}

	}

	public function equalizeTerrain(bitmap:BitmapData):void {
		// Adjust altitude histogram so that it's roughly quadratic and
		// water histogram so that it's roughly linear
		var histograms:Vector.<Vector.<Number>> = bitmap.histogram(bitmap.rect);
		var G:Vector.<Number> = histograms[1];
		var B:Vector.<Number> = histograms[2];
		var g:int = 0;
		var b:int = 0;
		var green:Array = new Array(256);
		var blue:Array = new Array(256);
		var cumsumG:Number = 0.0;
		var cumsumB:Number = 0.0;
		for (var i:int = 0; i < 256; i++) {
			cumsumG += G[i];
			cumsumB += B[i];
			green[i] = (g * g / 255) << 8; // int to green color value
			blue[i] = (b * b / 255); // int to blue color value
			while (cumsumG > SIZE * SIZE * Math.sqrt(g / 256.0) && g < 255) {
				g++;
			}
			while (cumsumB > SIZE * SIZE * (b / 256.0) && b < 255) {
				b++;
			}
		}
		bitmap.paletteMap(bitmap, bitmap.rect, new Point(0, 0), null, green, blue, null);

	}

	public function make2dArray(w:int, h:int):Vector.<Vector.<int>> {
		var v:Vector.<Vector.<int>> = new Vector.<Vector.<int>>(w);
		for (var x:int = 0; x < w; x++) {
			v[x] = new Vector.<int>(h);
			for (var y:int = 0; y < h; y++) {
				v[x][y] = 0;
			}
		}
		return v;
	}

}

